@@ -21,7 +21,8 @@ from sqlalchemy.types import \
     Boolean, BigInteger, Date, SmallInteger, Unicode, UnicodeText
 
 __all__ = [
-    'Medline', 'Author', 'Chemical', 'Database', 'Descriptor', 'Identifier', 'Qualifier', 'Section'
+    'Medline', 'Author', 'Chemical', 'Database', 'Descriptor',
+    'Identifier', 'Keyword', 'Qualifier', 'Section', 'PublicationType'
 ]
 
 _Base = declarative_base()
@@ -30,6 +31,7 @@ _session = lambda *args, **kwds: None
 
 NULL = lambda s: '\\N' if s is None else s
 DATE = lambda s: '\\N' if s is None else s.isoformat()
+STRING = lambda s: s.replace('\\', '\\\\').replace('\t', '\\t').replace('\n', '\\n')
 
 logger = logging.getLogger(__name__)
 
@@ -47,17 +49,18 @@ def InitDb(*args, **kwds):
     global _Base
     global _db
     global _session
+
     if len(args) > 0:
         # inject the foreign key pragma when using SQLite databases to ensure integrity
         # http://docs.sqlalchemy.org/en/rel_0_8/dialects/sqlite.html#foreign-key-support
         if (isinstance(args[0], str) and args[0].startswith('sqlite')) or \
                 (isinstance(args[0], URL) and args[0].get_dialect() == 'sqlite'):
-            #noinspection PyUnusedLocal
             @event.listens_for(engine.Engine, "connect")
             def set_sqlite_pragma(dbapi_connection, _):
                 cursor = dbapi_connection.cursor()
                 cursor.execute("PRAGMA foreign_keys=ON")
                 cursor.close()
+
     _db = engine.create_engine(*args, **kwds)
     _Base.metadata.create_all(_db)
     _session = session.sessionmaker(bind=_db)
@@ -106,7 +109,6 @@ class SelectMixin(object):
     (i.e., `Medline`, except for `Qualifier`).
     """
 
-    # noinspection PyUnresolvedReferences
     @classmethod
     def _buildQuery(cls, columns, parent_pk):
         if isinstance(parent_pk, tuple):
@@ -136,7 +138,6 @@ class SelectMixin(object):
         :arg  attributes: A list of attribute (column) names
         :type attributes: :class:`list` of :class:`str`
         """
-        # noinspection PyUnresolvedReferences
         mapping = dict((col.key, col) for col in cls.__table__.c)
         columns = [mapping[name] for name in attributes]
         query = cls._buildQuery(columns, parent_pk)
@@ -149,7 +150,6 @@ class SelectMixin(object):
 
         See :meth:`select` for details.
         """
-        # noinspection PyUnresolvedReferences
         query = cls._buildQuery([cls.__table__], parent_pk)
         return _fetch_all(query)
 
@@ -190,7 +190,7 @@ class Identifier(_Base, SelectMixin):
 
     def __str__(self):
         return '{}\t{}\t{}\n'.format(
-            NULL(self.pmid), NULL(self.namespace), NULL(self.value)
+            NULL(self.pmid), NULL(self.namespace), STRING(self.value)
         )
 
     def __repr__(self):
@@ -298,14 +298,14 @@ class Author(_Base, SelectMixin):
         UnicodeText, CheckConstraint("name <> ''"), nullable=False
     )
     initials = Column(Unicode(length=128), nullable=True)
-    forename = Column(Unicode(length=128), nullable=True)
+    forename = Column(Unicode(length=256), nullable=True)
     suffix = Column(Unicode(length=128), nullable=True)
 
     def __init__(self, pmid: int, pos: int, name: str,
                  initials: str=None, forename: str=None, suffix: str=None):
         assert pmid > 0, pmid
         assert pos > 0, pos
-        assert name
+        assert name, repr(name)
         self.pmid = pmid
         self.pos = pos
         self.name = name
@@ -315,7 +315,7 @@ class Author(_Base, SelectMixin):
 
     def __str__(self):
         return "{}\t{}\t{}\t{}\t{}\t{}\n".format(
-            NULL(self.pmid), NULL(self.pos), NULL(self.name),
+            NULL(self.pmid), NULL(self.pos), STRING(self.name),
             NULL(self.initials), NULL(self.forename), NULL(self.suffix))
 
     def __repr__(self):
@@ -368,10 +368,10 @@ class Qualifier(_Base, SelectMixin):
             the descriptor order in the record (starting from 1)
         sub
             the qualifier order within the descriptor (starting from 1)
-        name
-            the qualifier (name)
         major
             ``True`` if major, ``False`` if minor
+        name
+            the qualifier (name)
 
     Primary Key: ``(pmid, num, sub)``
     """
@@ -387,26 +387,24 @@ class Qualifier(_Base, SelectMixin):
     pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
     num = Column(SmallInteger, primary_key=True)
     sub = Column(SmallInteger, CheckConstraint("sub > 0"), primary_key=True)
-    name = Column(
-        UnicodeText, CheckConstraint("name <> ''"), nullable=False
-    )
     major = Column(Boolean, nullable=False)
+    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)
 
     def __init__(self, pmid: int, num: int, sub: int, name: str, major: bool=False):
         assert pmid > 0, pmid
         assert num > 0, num
         assert sub > 0, sub
-        assert name
+        assert name, repr(name)
         self.pmid = pmid
         self.num = num
         self.sub = sub
-        self.name = name
         self.major = major
+        self.name = name
 
     def __str__(self):
         return '{}\t{}\t{}\t{}\t{}\n'.format(
-            NULL(self.pmid), NULL(self.num), NULL(self.sub), NULL(self.name),
-            'T' if self.major else 'F'
+            NULL(self.pmid), NULL(self.num), NULL(self.sub),
+            'T' if self.major else 'F', STRING(self.name),
         )
 
     def __repr__(self):
@@ -431,10 +429,10 @@ class Descriptor(_Base, SelectMixin):
             the record this descriptor name belongs to
         num
             the descriptor order in the record (starting from 1)
-        name
-            the descriptor
         major
             ``True`` if major, ``False`` if minor
+        name
+            the descriptor
 
     Relations:
 
@@ -448,10 +446,8 @@ class Descriptor(_Base, SelectMixin):
 
     pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
     num = Column(SmallInteger, CheckConstraint("num > 0"), primary_key=True)
-    name = Column(
-        UnicodeText, CheckConstraint("name <> ''"), nullable=False
-    )
     major = Column(Boolean, nullable=False)
+    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)
 
     qualifiers = relation(
         Qualifier, backref='descriptor', cascade="all",
@@ -459,18 +455,17 @@ class Descriptor(_Base, SelectMixin):
     )
 
     def __init__(self, pmid: int, num: int, name: str, major: bool=False):
-        assert pmid > 0
-        assert num > 0
-        assert name
+        assert pmid > 0, pmid
+        assert num > 0, num
+        assert name, repr(name)
         self.pmid = pmid
         self.num = num
-        self.name = name
         self.major = major
+        self.name = name
 
     def __str__(self):
         return '{}\t{}\t{}\t{}\n'.format(
-            NULL(self.pmid), NULL(self.num), NULL(self.name),
-            'T' if self.major else 'F'
+            NULL(self.pmid), NULL(self.num), 'T' if self.major else 'F', STRING(self.name)
         )
 
     def __repr__(self):
@@ -492,49 +487,89 @@ class Chemical(_Base, SelectMixin):
 
         pmid
             the record's identifier (PubMed ID)
-        num
+        idx
             the order in the XML record
         uid
             the chemical's "unique identifier" (registry number, etc.)
         name
             the chemical's name
 
-    Primary Key: ``(pmid, num)``
+    Primary Key: ``(pmid, idx)``
     """
 
     __tablename__ = 'chemicals'
 
     pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
-    num = Column(SmallInteger, CheckConstraint("num > 0"), primary_key=True)
-    uid = Column(Unicode(length=256))
+    idx = Column(SmallInteger, CheckConstraint("idx > 0"), primary_key=True)
+    uid = Column(Unicode(length=256), CheckConstraint("uid <> ''"), nullable=True)
     name = Column(Unicode(length=256), CheckConstraint("name <> ''"), nullable=False)
 
-    def __init__(self, pmid: int, num: int, uid, name: str):
-        assert pmid > 0
-        assert num > 0
-        assert name
+    def __init__(self, pmid: int, idx: int, name: str, uid: str=None):
+        assert pmid > 0, pmid
+        assert idx > 0, idx
+        assert name, repr(name)
+        assert uid is None or uid, repr(uid)
         self.pmid = pmid
-        self.num = num
+        self.idx = idx
         self.uid = uid
         self.name = name
 
     def __str__(self):
         return '{}\t{}\t{}\t{}\n'.format(
-            NULL(self.pmid), NULL(self.num), NULL(self.uid), self.name
+            NULL(self.pmid), NULL(self.idx), NULL(self.uid), STRING(self.name)
         )
 
     def __repr__(self):
-        return "Chemical<{}:{}>".format(self.pmid, self.num)
+        return "Chemical<{}:{}>".format(self.pmid, self.idx)
 
     def __eq__(self, other):
         return isinstance(other, Chemical) and \
                self.pmid == other.pmid and \
-               self.num == other.num
+               self.idx == other.idx and \
+               self.uid == other.uid and \
+               self.name == other.name
+
+
+class PublicationType(_Base, SelectMixin):
+    """
+    Records the type of publication this citation refers to.
+
+    Attributes:
+
+        pmid
+            the record's identifier (PubMed ID)
+        value
+            the publication type value
+
+    Primary Key: ``(pmid, value)``
+    """
+
+    __tablename__ = 'publication_types'
+
+    pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
+    value = Column(Unicode(length=256), CheckConstraint("value <> ''"), primary_key=True)
+
+    def __init__(self, pmid: int, value: str):
+        assert pmid > 0, pmid
+        assert value, repr(value)
+        self.pmid = pmid
+        self.value = value
+
+    def __str__(self):
+        return '{}\t{}\n'.format(NULL(self.pmid), NULL(self.value))
+
+    def __repr__(self):
+        return "PublicationType<{}:{}>".format(self.pmid, self.value)
+
+    def __eq__(self, other):
+        return isinstance(other, PublicationType) and \
+               self.pmid == other.pmid and \
+               self.value == other.value
 
 
 class Database(_Base, SelectMixin):
     """
-    References to external databases curated my the NLM.
+    References to external databases curated by the NLM.
 
     Attributes:
 
@@ -555,8 +590,8 @@ class Database(_Base, SelectMixin):
     accession = Column(Unicode(length=256), CheckConstraint("accession <> ''"), primary_key=True)
 
     def __init__(self, pmid: int, name: str, accession: str):
-        assert pmid > 0
-        assert name
+        assert pmid > 0, pmid
+        assert name, repr(name)
         assert accession, repr(accession)
         self.pmid = pmid
         self.name = name
@@ -564,7 +599,7 @@ class Database(_Base, SelectMixin):
 
     def __str__(self):
         return '{}\t{}\t{}\n'.format(
-            NULL(self.pmid), NULL(self.name), NULL(self.accession)
+            NULL(self.pmid), NULL(self.name), STRING(self.accession)
         )
 
     def __repr__(self):
@@ -577,6 +612,68 @@ class Database(_Base, SelectMixin):
                self.accession == other.accession
 
 
+class Keyword(_Base, SelectMixin):
+    """
+    Keywords, external or curated by the NLM.
+
+    Attributes:
+
+        pmid
+            the record's identifier (PubMed ID)
+        owner
+            the entity that provided the keyword
+        cnt
+            a unique counter for all keywords from a given owner and record
+            (starting from 1)
+        major
+            if the keyword is a major topic of this article
+        name
+            the keyword itself
+
+    Primary Key: ``(pmid, owner, cnt)``
+    """
+
+    __tablename__ = 'keywords'
+
+    OWNERS = frozenset({'NASA', 'PIP', 'KIE', 'NLM', 'NOTNLM', 'HHS'})
+
+    pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
+    owner = Column(Enum(*OWNERS, name='owner'), primary_key=True)
+    cnt = Column(SmallInteger, CheckConstraint("cnt > 0"), primary_key=True)
+    major = Column(Boolean, nullable=False)
+    name = Column(UnicodeText, CheckConstraint("name <> ''"), nullable=False)
+
+    def __init__(self, pmid: int, owner: str, cnt: int, name: str, major: bool=False):
+        assert pmid > 0, pmid
+        assert owner in Keyword.OWNERS, repr(owner)
+        assert cnt > 0, cnt
+        assert name, repr(name)
+        self.pmid = pmid
+        self.owner = owner
+        self.cnt = cnt
+        self.major = major
+        self.name = name
+
+    def __str__(self):
+        return '{}\t{}\t{}\t{}\t{}\n'.format(
+            NULL(self.pmid), NULL(self.owner), NULL(self.cnt),
+            'T' if self.major else 'F', STRING(self.name)
+        )
+
+    def __repr__(self):
+        return "Keyword<{}:{}:{}>".format(
+                self.pmid, self.owner, self.cnt
+        )
+
+    def __eq__(self, other):
+        return isinstance(other, Keyword) and \
+               self.pmid == other.pmid and \
+               self.owner == other.owner and \
+               self.cnt == other.cnt and \
+               self.major == other.major and \
+               self.name == other.name
+
+
 class Section(_Base, SelectMixin):
     """
     The text sections of the records.
@@ -588,34 +685,28 @@ class Section(_Base, SelectMixin):
         seq
             the sequence of sections in the record (starting from 1)
         name
-            the name of the section (see `Section.SECTIONS`)
+            the name of the section (Title, Abstract, Vernacular, Copyright, ...)
         label
             section label as defined by the publisher (if any)
         content
             the text content of this section
 
-    Primary Key: ``(pmid, num)``
+    Primary Key: ``(pmid, seq)``
     """
-
-    SECTIONS = frozenset({'Title', 'Abstract', 'Vernacular', 'Copyright',
-                          'Background', 'Objective', 'Methods', 'Results', 'Conclusions',
-                          'Unassigned', 'Unlabelled'})
-
     __tablename__ = 'sections'
 
     pmid = Column(BigInteger, ForeignKey('records', ondelete="CASCADE"), primary_key=True)
     seq = Column(SmallInteger, CheckConstraint("seq > 0"), primary_key=True)
-    name = Column(Enum(*SECTIONS, name='section'), nullable=False)
-    label = Column(Unicode(length=256))
-    content = Column(
-        UnicodeText, CheckConstraint("content <> ''"), nullable=False
-    )
+    name = Column(Unicode(length=64), CheckConstraint("name <> ''"), nullable=False)
+    label = Column(Unicode(length=256), CheckConstraint("label <> ''"), nullable=True)
+    content = Column(UnicodeText, CheckConstraint("content <> ''"), nullable=False)
 
     def __init__(self, pmid: int, seq: int, name: str, content: str, label: str=None):
-        assert pmid > 0
-        assert seq > 0
-        assert name in Section.SECTIONS, name
-        assert content
+        assert pmid > 0, pmid
+        assert seq > 0, seq
+        assert name, repr(name)
+        assert content, repr(content)
+        assert label is None or label, repr(label)
         self.pmid = pmid
         self.seq = seq
         self.name = name
@@ -625,9 +716,7 @@ class Section(_Base, SelectMixin):
     def __str__(self):
         return '{}\t{}\t{}\t{}\t{}\n'.format(
             NULL(self.pmid), NULL(self.seq), NULL(self.name),
-            NULL(self.label), NULL(
-                self.content.replace('\\', '\\\\').replace('\t', '\\t').replace('\n', '\\n')
-            )
+            NULL(self.label), STRING(self.content)
         )
 
     def __repr__(self):
@@ -694,64 +783,76 @@ class Medline(_Base):
 
     __tablename__ = 'records'
 
-    sections = relation(
-        Section, backref='medline', cascade='all, delete-orphan',
-        order_by=Section.__table__.c.seq
-    )
     authors = relation(
         Author, backref='medline', cascade='all, delete-orphan',
         order_by=Author.__table__.c.pos
     )
-    identifiers = relation(
-        Identifier, backref='medline', cascade='all, delete-orphan',
-        collection_class=column_mapped_collection(Identifier.namespace)
+    chemicals = relation(
+        Chemical, backref='medline', cascade='all, delete-orphan',
+        order_by=Chemical.__table__.c.idx
     )
     databases = relation(
         Database, backref='medline', cascade='all, delete-orphan',
     )
-    chemicals = relation(
-        Chemical, backref='medline', cascade='all, delete-orphan',
-        order_by=Chemical.__table__.c.num
-    )
     descriptors = relation(
         Descriptor, backref='medline', cascade='all, delete-orphan',
         order_by=Descriptor.__table__.c.num
     )
+    identifiers = relation(
+        Identifier, backref='medline', cascade='all, delete-orphan',
+        collection_class=column_mapped_collection(Identifier.namespace)
+    )
+    keywords = relation(
+        Keyword, backref='medline', cascade='all, delete-orphan',
+        order_by=(Keyword.__table__.c.owner, Keyword.__table__.c.cnt)
+    )
+    publication_types = relation(PublicationType, backref='medline', cascade='all, delete-orphan')
     qualifiers = relation(Qualifier, backref='medline')
-
-    pmid = Column(BigInteger, CheckConstraint('pmid > 0'),
-                  primary_key=True, autoincrement=False)
-    status = Column(Enum(*STATES, name='state'), nullable=False)
-    journal = Column(Unicode(length=256), CheckConstraint("journal <> ''"),
-                     nullable=False)
-    created = Column(Date, nullable=False)
-    completed = Column(Date)
-    revised = Column(Date)
-    modified = Column(
-        Date, default=date.today, onupdate=date.today, nullable=False
+    sections = relation(
+        Section, backref='medline', cascade='all, delete-orphan',
+        order_by=Section.__table__.c.seq
     )
 
-    def __init__(self, pmid: int, status: str, journal: str,
-                 created: date, completed: date=None, revised: date=None):
-        assert pmid > 0
-        assert status in Medline.STATES, status
-        assert journal
-        assert isinstance(created, date)
-        assert completed is None or isinstance(completed, date)
-        assert revised is None or isinstance(revised, date)
+    pmid = Column(BigInteger, CheckConstraint('pmid > 0'), primary_key=True, autoincrement=False)
+    status = Column(Enum(*STATES, name='state'), nullable=False)
+    journal = Column(Unicode(length=256), CheckConstraint("journal <> ''"), nullable=False)
+    pub_date = Column(Unicode(length=256), CheckConstraint("pub_date <> ''"), nullable=False)
+    issue = Column(Unicode(length=256), CheckConstraint("issue <> ''"), nullable=True)
+    pagination = Column(Unicode(length=256), CheckConstraint("pagination <> ''"), nullable=True)
+    created = Column(Date, nullable=False)
+    completed = Column(Date, nullable=True)
+    revised = Column(Date, nullable=True)
+    modified = Column(Date, default=date.today, onupdate=date.today, nullable=False)
+
+    def __init__(self, pmid: int, status: str, journal: str, pub_date: str,
+                 created: date, completed: date=None, revised: date=None,
+                 issue: str=None, pagination: str=None):
+        assert pmid > 0, pmid
+        assert status in Medline.STATES, repr(status)
+        assert journal, repr(journal)
+        assert pub_date, repr(pub_date)
+        assert isinstance(created, date), repr(created)
+        assert completed is None or isinstance(completed, date), repr(completed)
+        assert revised is None or isinstance(revised, date), repr(revised)
+        assert pagination is None or pagination
+        assert issue is None or issue
         self.pmid = pmid
         self.status = status
         self.journal = journal
+        self.pub_date = pub_date
+        self.issue = issue
+        self.pagination = pagination
         self.created = created
         self.completed = completed
         self.revised = revised
 
     def __str__(self):
-        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(
+        return '{}\n'.format('\t'.join(map(str, [
             NULL(self.pmid), NULL(self.status), NULL(self.journal),
+            NULL(self.pub_date), NULL(self.issue), NULL(self.pagination),
             DATE(self.created), DATE(self.completed), DATE(self.revised),
             DATE(date.today() if self.modified is None else self.modified)
-        )
+        ])))
 
     def __repr__(self):
         return "Medline<{}>".format(self.pmid)
@@ -761,10 +862,18 @@ class Medline(_Base):
                self.pmid == other.pmid and \
                self.status == other.status and \
                self.journal == other.journal and \
+               self.pub_date == other.pub_date and \
+               self.issue == other.issue and \
+               self.pagination == other.pagination and \
                self.created == other.created and \
                self.completed == other.completed and \
                self.revised == other.revised
 
+    def citation(self):
+        issue = '; {}'.format(self.issue) if self.issue else ""
+        pagination = ': {}'.format(self.pagination) if self.pagination else ""
+        return "{}{}{}".format(self.pub_date, issue, pagination)
+
     @classmethod
     def insert(cls, data: dict):
         """
